/*

  FLASH.C  --  program to display Chinese flashcards

  Version 1.5  --  September 22, 1993

  This program copyright (C) Stephen G. Simpson <simpson@math.psu.edu>

  To compile with Turbo C 2.0, do ``TCC FLASH HBF GRAPHICS.LIB''.

*/

#include <graphics.h>
#include <stdio.h>
#include <stdlib.h>
#include "hbf.h"

#define SFONTENV "SFONT"     /* environment name for simplified font */
#define TFONTENV "TFONT"     /* environment name for traditional font */
#define SFONTDEFAULT "SIMP24.HBF"   /* default simplified font */
#define TFONTDEFAULT "TRAD24.HBF"   /* default traditional font */
#define MAXFILEPATH 160      /* maximum length of a file path */

#define WIDTH 24          /* width of a bitmap */
#define HEIGHT 24         /* height of a bitmap */
#define XALL 3            /* x-coordinate for displaying everything */
#define YTRAD 1           /* y-coordinate for traditional characters */
#define YSIMP 63          /* y-coordinate for simplified characters */
#define YPHON 121         /* y-coordinate for phonetic string */
#define YENGL 150         /* y-coordinate for English meaning */
#define YEOF 186          /* y-coordinate for - and + markers */
#define MAXLINE 120       /* maximum length of a line in a flashcard file */
#define MAXCHARS 10       /* maximum number of characters in a phrase */
#define MAXCARDS 100      /* maximum number of cards to store in memory */
#define COMMENT '#'       /* ASCII byte used to start a comment line */

static int color;
static char* progdir;
static HBF *fontS;        /* font for simplified characters */
static HBF *fontT;        /* font for traditional characters */
static void startgraphics();
static void endgraphics();
struct ccode {
  unsigned char high, low;
};
struct flashcard {
  struct ccode trad [MAXCHARS+1], simp [MAXCHARS+1];
  unsigned char phon [MAXLINE+1], engl [MAXLINE+1];
};
static int getline(unsigned char* line, int max, FILE*);
static void parse(struct flashcard*, unsigned char* line);
static void display(struct flashcard*, char* displaycontrolstring);
static void displayS(int x, int y, struct ccode);
static void displayT(int x, int y, struct ccode);
static void displaybitmap(int x, int y,
  unsigned char* bitmap, int width, int height);
static void displaybox(int x, int y, int width, int height);

void main(int argc, char *argv[]) {
  char* progname;
  FILE* file;
  unsigned char line[MAXLINE+3];
  struct flashcard card[MAXCARDS];
  int key, responded, reviewing, i, p, n, isline;

  i = strlen(argv[0]);
  while (argv[0][i] != '.') i--; argv[0][i] = '\0';
  while (argv[0][i] != '\\') i--; argv[0][i] = '\0';
  progdir = argv[0]; progname = argv[0] + i + 1;
  if (argc == 2 && (argv[1][0] == 'h' || argv[1][0] == 'H')) {
    printf("\nCommand syntax: %s XXX YYY <file>\n\n", progname);
    printf("where XXX and YYY consist of one or more display ");
    printf("control letters,\n");
    printf("and <file> is a Chinese flashcard file.  XXX controls ");
    printf("the query and\nYYY controls the response.  ");
    printf("The recognized display control letters\nare:\n\n");
    printf("T or C -- traditional Chinese characters\n");
    printf("P -- pronunciation\n");
    printf("E -- English meaning\n");
    printf("S -- simplified Chinese characters\n");
    printf("U -- simplified characters if different from traditional\n");
    printf("V -- traditional characters if different from simplified\n");
    printf("N -- does nothing\n\n");
    printf("Example: %s C PE CARDFILE\n\n", progname);
    printf("In this example, the query consists of a Chinese character ");
    printf("or phrase,\nand the response consists of the pronunciation ");
    printf("and English meaning.\nTo display all of the queries and ");
    printf("responses in order, press the\nspace bar repeatedly.\n");
    exit(0);
  }
  if (argc != 4) {
    printf("Syntax: %s XXX YYY <file>\n", progname);
    printf("or, for more help, %s HELP\n", progname);
    exit(1);
  }
  if ((file = fopen(argv[3], "rb")) == NULL) {
    fclose(file); printf("Error: can't open file: %s\n", argv[3]); exit(1);
  }
  /* read the first card line from the file */
  isline = getline(line, MAXLINE, file);
  /* if there are no card lines, terminate the program */
  if (!isline) {
    printf("Error: no cards in file: %s\n", argv[3]); exit(1);
  }
  /* otherwise, display the cards in order */
  startgraphics();
  /*  startgraphics sets default colors and text style */
  /*  but you can change them here */
  /*  color = YELLOW; */    /* color used by displaybitmap and displaybox */
  /*  setcolor(BROWN); */   /* color used by BGI outtext functions */
  /*  settextstyle(TRIPLEX_FONT, HORIZ_DIR, 3); */
  settextstyle(SANS_SERIF_FONT, HORIZ_DIR, 3);
  p = -1; n = 0;
  while (isline) {
    /* there is a card line waiting to be stored as a card */
    p++; p %= MAXCARDS;
    /* p is where to store the card */
    parse(&card[p], line);
    /* p is where the card was stored */
    if (n < MAXCARDS) n++;
    /* n is the total number of cards that are currently stored */
    /* look ahead to find out whether */
    /* there are any more card lines in the file */
    isline = getline(line, MAXLINE, file);
    /* now display some cards, beginning with the one that was just stored */
    for (i = p, reviewing = 1; reviewing; ) {
      /* i is the card to be displayed */
      cleardevice();
      if (n > 1 && i == (p+1) % n) outtextxy(XALL, YEOF, "-");
      if (i == p && !isline) outtextxy(XALL, YEOF, "+");
      /* display the query */
      display(&card[i], argv[1]);
      responded = 0;
      /* get keyboard input and act accordingly */
      while (1) {
        key = getch();
        if (key == '>' || key == '.' || key == '+'
            || (key == ' ' && responded)) {
          if (i == p) reviewing = 0;
          i = (i+1) % n;
          break;
        }
        if (key == ' ' && !responded) {
          /* display the response */
          display(&card[i], argv[2]);
          responded = 1;
        }
        if (key == '<' || key == ',' || key == '-' || key == '\010') {
          if (n > 1 && i != (p+1) % n) {
            if (i > 0) i--; else i = n-1;
            reviewing = 1;
            break;
          }
          if (n == 1) outtextxy(XALL, YEOF, "-");
        }
        if (key == '\r') {
           if (i == p) reviewing = 0;
           i = p;
           break;
        }
        if (key == '\033' || key == 'q' || key == 'Q') {
          fclose(file); endgraphics(); exit(0);
        }
      }
    }
  }
  /* terminate the program */
  fclose(file); endgraphics(); exit(0);
}

static int getline(unsigned char* line, int max, FILE* file) {
  /* line should have max + 3 bytes allocated */
  int len, c;

  for (len = 0; len == 0; ) {
    if (!fgets(line, max+3, file)) break;
    len = strlen(line);
    if (len > 0 && line[len-1] != '\n')
      while ((c = fgetc(file)) !=  '\n' && c != EOF) ;
    if (line[0] == COMMENT) len = 0;
    if (len > max) len = max;
    if (len > 0 && ((c = line[len-1]) == '\n' || c == '\r')) len--;
    if (len > 0 && ((c = line[len-1]) == '\n' || c == '\r')) len--;
    line[len] = '\0';
  }
  return(len);
}

static void startgraphics() {
  char spath [MAXFILEPATH];
  char tpath [MAXFILEPATH];
  char *sfont, *tfont, c;
  int gdriver, gmode, gerror, n;
  FILE* f;

  /* initialize simplified character font */
  if ((sfont = getenv(SFONTENV)) == NULL) sfont = SFONTDEFAULT;
  if (sfont[0] == '\\' || sfont[1] == ':')
    strncpy(spath, sfont, MAXFILEPATH - 1);
  else {
    strncpy(spath, progdir, MAXFILEPATH - 2);
    spath [MAXFILEPATH - 2] = '\0';
    c = spath [(n = strlen(spath)) - 1];
    if (c != '\\' && c != ':') strcat(spath, "\\");
    strncat(spath, sfont, MAXFILEPATH - n - 2);
  }
  spath [MAXFILEPATH - 1] = '\0';
  if ((fontS = hbfOpen(spath)) == NULL) {
    printf("Error: invalid SFONT:\n%s\n", spath);
    exit(1);
  }
  if ((hbfBitmapBBox(fontS)->hbf_width != WIDTH) ||
      (hbfBitmapBBox(fontS)->hbf_height != HEIGHT)) {
    printf("Error: SFONT is not a %dx%d font:\n%s\n",
                   WIDTH, HEIGHT, spath);
    exit(1);
  }

  /* initialize traditional character font */
  if ((tfont = getenv(TFONTENV)) == NULL) tfont = TFONTDEFAULT;
  if (tfont[0] == '\\' || tfont[1] == ':')
    strncpy(tpath, tfont, MAXFILEPATH - 1);
  else {
    strncpy(tpath, progdir, MAXFILEPATH - 2);
    tpath [MAXFILEPATH - 2] = '\0';
    c = tpath [(n = strlen(tpath)) - 1];
    if (c != '\\' && c != ':') strcat(tpath, "\\");
    strncat(tpath, tfont, MAXFILEPATH - n - 2);
  }
  tpath [MAXFILEPATH - 1] = '\0';
  if ((fontT = hbfOpen(tpath)) == NULL) {
    printf("Error: invalid TFONT:\n%s\n", tpath);
    exit(1);
  }
  if ((hbfBitmapBBox(fontT)->hbf_width != WIDTH) ||
      (hbfBitmapBBox(fontT)->hbf_height != HEIGHT)) {
    printf("Error: TFONT is not a %dx%d font:\n%s\n",
                   WIDTH, HEIGHT, tpath);
    exit(1);
  }

  /* initialize the Borland graphics system */
  /*
  First, register the BGI drivers and fonts to be linked in.
  This assumes that they have been converted to .OBJ files
  using BGIOBJ /F and then added to GRAPHICS.LIB using TLIB.
  */
  registerfarbgidriver(CGA_driver_far);
  registerfarbgidriver(ATT_driver_far);
  registerfarbgidriver(Herc_driver_far);
  registerfarbgidriver(EGAVGA_driver_far);
  registerfarbgidriver(PC3270_driver_far);
  registerfarbgidriver(IBM8514_driver_far);
  registerfarbgifont(sansserif_font_far);
  /* registerfarbgifont(triplex_font_far); */
  /* registerfarbgifont(gothic_font_far); */
  /* registerfarbgifont(small_font_far); */
  /*
  Next, detect the graphics hardware and start the graphics system.
  */
  detectgraph(&gdriver, &gmode);
  initgraph(&gdriver, &gmode, NULL);
  if ((gerror = graphresult()) != grOk) {
    closegraph();
    printf("Graphics error: %s.\n", grapherrormsg(gerror));
    exit(1);
  }
  graphdefaults();
  color = getcolor();    /* color used by displaybitmap and displaybox */
}

static void endgraphics() {
  hbfClose(fontS);
  hbfClose(fontT);
  cleardevice();
  closegraph();
}

static void parse(struct flashcard* card, unsigned char* line) {
  int i, j;

  i = 0;
  while (line[i] == ' ' || line[i] == '\t') i++;
  for (j = 0; j < MAXCHARS; j++) {
    if (line[i] == ' ' || line[i] == '\t' || !line[i]) break;
    card->trad[j].high = line[i]; i++;
    if (line[i] == ' ' || line[i] == '\t' || !line[i]) break;
    card->trad[j].low = line[i]; i++;
  }
  card->trad[j].high = '\0';
  while (line[i] != ' ' && line[i] != '\t') {
    if (!line[i]) break; i++;
  }
  while (line[i] == ' ' || line[i] == '\t') i++;
  for (j = 0; j < MAXCHARS; j++) {
    if (line[i] == ' ' || line[i] == '\t' || !line[i]) break;
    card->simp[j].high = line[i]; i++;
    if (line[i] == ' ' || line[i] == '\t' || !line[i]) break;
    card->simp[j].low = line[i]; i++;
  }
  card->simp[j].high = '\0';
  while (line[i] != ' ' && line[i] != '\t') {
    if (!line[i]) break; i++;
  }
  while (line[i] == ' ' || line[i] == '\t') i++;
  for (j = 0; ; i++, j++) {
    if (line[i] == ' ' || line[i] == '\t' || !line[i]) break;
    if (line[i] == '_') card->phon[j] = ' ';
    else card->phon[j] = line[i];
  }
  card->phon[j] = '\0';
  while (line[i] == ' ' || line[i] == '\t') i++;
  for (j = 0; line[i]; i++, j++) card->engl[j] = line[i];
  card->engl[j] = '\0';
}

static void display(struct flashcard* card, char* string) {
  int i, j;

  for (i = 0; string[i]; i++) {
    switch(string[i]) {
      case 'c':
      case 'C':
      case 't':
      case 'T':
        for (j = 0; card->trad[j].high; j++) {
          if (card->trad[j].high == '*')
            displayS(XALL + 53*j, YTRAD, card->simp[j]);
          else displayT(XALL + 53*j, YTRAD, card->trad[j]);
        };
        break;
      case 'v':
      case 'V':
        for (j = 0; card->trad[j].high; j++)
          if (card->trad[j].high != '*' && card->simp[j].high != '*')
            displayT(XALL + 53*j, YTRAD, card->trad[j]);
        break;
      case 's':
      case 'S':
        for (j = 0; card->simp[j].high; j++) {
          if (card->simp[j].high == '*')
            displayT(XALL + 53*j, YSIMP, card->trad[j]);
          else displayS(XALL + 53*j, YSIMP, card->simp[j]);
        };
        break;
      case 'u':
      case 'U':
        for (j = 0; card->simp[j].high; j++)
          if (card->simp[j].high != '*' && card->trad[j].high != '*')
            displayS(XALL + 53*j, YSIMP, card->simp[j]);
        break;
      case 'p':
      case 'P': outtextxy(XALL, YPHON, card->phon); break;
      case 'e':
      case 'E': outtextxy(XALL, YENGL, card->engl); break;
      case '-':
      case '/':
      case 'x':
      case 'X':
      case 'n':
      case 'N': break;
      default:
        endgraphics();
        printf("Error: unknown display control letter: %c\n", string[i]);
        printf("in display control string: %s\n", string);
        exit(1);
    }
  }
}

static void displayS(int x0, int y0, struct ccode cc) {
  unsigned char* bitmap;
  unsigned int code;

  code = cc.high * 256 + cc.low;
  if ((bitmap = hbfGetBitmap(fontS, code)) != NULL)
     displaybitmap(x0, y0, bitmap, WIDTH, HEIGHT);
  else
     displaybox(x0, y0, WIDTH, HEIGHT);
}

static void displayT(int x0, int y0, struct ccode cc) {
  unsigned char* bitmap;
  unsigned int code;

  code = cc.high * 256 + cc.low;
  if ((bitmap = hbfGetBitmap(fontT, code)) != NULL)
     displaybitmap(x0, y0, bitmap, WIDTH, HEIGHT);
  else
     displaybox(x0, y0, WIDTH, HEIGHT);
}

static void displaybitmap(int x0, int y0,
  unsigned char* buf, int w, int h) {
  int u, x, y;
  unsigned char bit[8] = {0x80, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2, 0x1};

  u = (w+7)/8; /* the size of each row, measured in bytes */
  /* we are assuming that each row is padded to full bytes */
  /* the size of the entire bitmap, measured in bytes, is ((w+7)/8)*h */
  for (x = 0; x < w; x++) for (y = 0; y < h; y++)
    if (buf[(u*y)+(x/8)] & bit[x%8]) putpixel(x0+2*x, y0+2*y, color);
}

static void displaybox(int x0, int y0, int w, int h) {
  int x, y;

  for (x = 0; x < w; x++) {
    putpixel(x0+2*x, y0, color);
    putpixel(x0+2*x, y0+2*(h-1), color);
  }
  for (y = 1; y < h-1; y++) {
    putpixel(x0, y0+2*y, color);
    putpixel(x0+2*(h-1), y0+2*y, color);
  }
}

